from copy import deepcopy
from threading import Lock, Thread
from queue import Empty
import multiprocessing as mp
import socket
import pickle as pkl
import logging
import time
import numpy as np

from cognifly.utils.udp_interface import UDPInterface
from cognifly.utils.tcp_video_interface import TCPVideoInterface
from cognifly.utils.ip_tools import extract_ip, get_free_port
from cognifly.utils.functions import clip, get_angle_sequence_rad


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


# Constants for the school easy API

EASY_API_DEFAULT_SPEED = 0.2  # (m/s)
EASY_API_DEFAULT_YAW_RATE = 0.5  # (rad/s)
EASY_API_ADDITIONAL_DURATION = 10.0  # (s)
EASY_API_TAKEOFF_ALTITUDE = 0.5  # should be roughly the same as default takeoff altitude (m)
EASY_API_MAX_ALTITUDE = 0.9
EASY_API_MIN_ALTITUDE = 0.35


# Global GUI thread (only one GUI is available for all drones)


def _gui_process(name, recv_q, send_q):
    """
    Process in charge of managing the Graphical User Interface

    This process communicates with the GUI thread of the Cognifly object

    Args:
        name: str: name of the window
        recv_q: mutliprocessing.Queue: the process receives tuples of the form ("CMD", value) here
        send_q: mutliprocessing.Queue: the process sends tuples of the form ("CMD", value) here
    """
    import PySimpleGUI as sg
    import cv2

    sg.theme('BluePurple')
    layout = [[sg.Text('Battery:'), sg.Text('OK', size=(15, 1), key='-batt-')],
              [sg.Button('DISARM')],
              [sg.Image(filename='', key='-image-')]]
    window = sg.Window(name,
                       layout,
                       keep_on_top=True,
                       enable_close_attempted_event=True)
    image_elem = window['-image-']
    batt_elem = window['-batt-']

    try:
        while True:  # Event Loop
            # when nothing has been happened, we sleep for a while at the end of the loop:
            no_window_event = False
            no_read_event = False

            # check window events:
            event, values = window.read(0)
            if event == 'DISARM':  # disarm button pressed
                send_q.put(('ARM', False))
            elif event == sg.WINDOW_CLOSE_ATTEMPTED_EVENT and sg.popup_yes_no(
                    'You will not be able to open another GUI.\nDo you really want to exit?',
                    keep_on_top=True) == 'Yes':
                send_q.put(('CLOSE', None))
                break
            else:
                no_window_event = True

            # check recv_q:
            try:
                cmd, val = recv_q.get_nowait()
            except Empty:
                cmd, val = None, None
                no_read_event = True

            if cmd is not None:
                if cmd == 'IMG':
                    if val is None:  # stop display
                        image_elem.update(data=b'')
                    else:  # display frame
                        frame = val
                        imgbytes = cv2.imencode('.png', frame)[1].tobytes()
                        image_elem.update(data=imgbytes)
                elif cmd == 'BAT':  # batt str value
                    batt_str = val
                    batt_elem.update(batt_str)

            if no_window_event and no_read_event:
                time.sleep(0.05)
    finally:
        # wait for closing acknowledgement:
        while True:
            try:
                cmd, _ = recv_q.get(block=True, timeout=0.1)
            except Empty:
                break
            if cmd == "CLOSE_ACK":
                break
        window.close()


# Remote controller class

class Cognifly:
    """
    Remote UDP controller meant to communicate with a drone that executes cognifly_controller.py
    Each command is sent via UDP
    The command is sent along an identifier that is sent back by the drone as an aknowledgement
    If the acknowledgement of the last sent command is not received after wait_ack_duration seconds, the last command is sent again
    Any command received by the drone overrides the previous command instantaneously
    """
    def __init__(self,
                 drone_hostname,
                 local_hostname=None,
                 send_port=8988,
                 recv_port=8989,
                 recv_port_video=8990,
                 wait_for_first_obs=True,
                 wait_for_first_obs_sleep_duration=0.1,
                 wait_ack_duration=0.2,
                 gui=True):
        """
        Args:
            drone_hostname: string: name of the drone (can be an ip address).
            local_hostname: string: name of the local machine (can be an ip address). If None, this is retrieved automatically.
            send_port: int: udp_send_port for the local UDPInterface.
            recv_port: int: udp_recv_port for UDPInterface.
            wait_for_first_obs: int: If True, waits for a first observation from the drone on isntantiation.
            wait_for_first_obs_sleep_duration: float: used only if wait_for_first_obs is True
            wait_ack_duration: float: If > 0.0, the framework will wait this number of seconds for an acknowledgement before resending.
                If <= 0.0, the resending feature is disabled.
            gui: bool: if True, a gui window will pop when the object is created (emergency disarm, battery, streaming).
        """
        self.gui = gui
        self.drone_hostname = drone_hostname
        self.send_port = send_port
        self.recv_port = recv_port
        self.recv_port_video = recv_port_video
        self.wait_for_first_obs = wait_for_first_obs
        self.wait_for_first_obs_sleep_duration = wait_for_first_obs_sleep_duration
        self.ea_speed = EASY_API_DEFAULT_SPEED
        self.ea_yaw_rate = EASY_API_DEFAULT_YAW_RATE

        self.udp_int = UDPInterface()
        self.tcp_video_int = TCPVideoInterface()

        self.local_hostname = socket.gethostname() if local_hostname is None else local_hostname
        self.local_ip = socket.gethostbyname(self.local_hostname) if local_hostname is not None else extract_ip()
        logger.info(f"local hostname: {self.local_hostname} with IP: {self.local_ip}")
        self.drone_hostname = drone_hostname
        self.drone_ip = socket.gethostbyname(drone_hostname)
        logger.info(f"drone hostname: {self.drone_hostname} with IP: {self.drone_ip}")

        self.udp_int.init_sender(self.drone_ip, self.send_port)
        free_port = get_free_port(self.recv_port)
        assert free_port is not None, "No available port!"
        if free_port != self.recv_port:
            logger.warning(f"Port {self.recv_port} is unavailable, trying to communicate on port {free_port} instead.")
            self.recv_port = free_port
        self.udp_int.init_receiver(self.local_ip, self.recv_port)

        self.wait_ack_duration = wait_ack_duration
        self.last_im_id = -1

        self.easy_api_cur_z = 0.0
        self.time_takeoff = time.time()

        self._lock = Lock()  # used to update the observation and also to resend
        self.__obs = None
        self.__last_sent_id = 0
        self.__last_received_id = 0
        self.__last_timestamp = time.time()
        self.__last_sent = None
        self.__wait_ack_reset = True
        self.__last_i_obs = 0
        self.__wait_done = False

        self._gui_lock = Lock()
        self._gui_display = False
        self._gui_batt_str = "OK"
        self._gui_process = None

        self._listener_thread = Thread(target=self.__listener_thread)
        self._listener_thread.setDaemon(True)  # thread will be terminated at exit
        self._listener_thread.start()
        if self.wait_ack_duration > 0:
            self._resender_thread = Thread(target=self.__resender_thread)
            self._resender_thread.setDaemon(True)  # thread will be terminated at exit
            self._resender_thread.start()
        else:
            self._resender_thread = None

        self.reset()

        with self._gui_lock:
            if self.gui:
                _gui_thread = Thread(target=self.__gui_thread, args=(self.drone_hostname, ))
                _gui_thread.setDaemon(True)  # thread will be terminated at exit
                _gui_thread.start()

        self.streamoff()

    def __gui_thread(self, gui_name):
        """
        Thread in charge of managing the Graphical User Interface

        This thread sends to and receives from the GUI process

        Args:
            gui_name: str: the name of the GUI window
        """
        mp_ctx = mp.get_context('spawn')  # to ensure the 'spawn' context is used on Linux
        send_q = mp_ctx.Queue()  # recv_q of the process
        recv_q = mp_ctx.Queue()  # send_q of the process
        with self._gui_lock:
            self._gui_process = mp_ctx.Process(target=_gui_process, args=(gui_name, send_q, recv_q))
            self._gui_process.start()
        is_displaying = False
        previous_batt_str = "OK"
        while True:
            send_event = False
            recv_event = False
            # commands from the process:
            try:
                cmd, val = recv_q.get_nowait()
            except Empty:
                cmd, val = None, None
            if cmd is not None:
                recv_event = True
                if cmd == 'ARM':
                    if not val:
                        self.disarm()
                    else:
                        self.arm()
                elif cmd == "CLOSE":
                    send_q.put(('CLOSE_ACK', None))
                    with self._gui_lock:
                        self._gui_process.join()  # wait for process end before we close the thread
                    break
            # display:
            with self._gui_lock:
                gui_display = self._gui_display
            if gui_display:
                send_event = True  # we don't want to wait if display is on, since there is a blocking call
                is_displaying = True
                with self.tcp_video_int.condition_in:
                    if self.tcp_video_int.condition_in.wait(timeout=0.1):  # wait for new frame
                        frame, _ = self.tcp_video_int.get_last_image()
                    else:
                        frame = None
                if frame is not None:
                    send_q.put(('IMG', frame))
            elif is_displaying:
                send_q.put(('IMG', None))
                send_event = True
                is_displaying = False
            # battery:
            with self._gui_lock:
                batt_str = self._gui_batt_str
            if batt_str != previous_batt_str:
                send_q.put(('BAT', batt_str))
                send_event = True
                previous_batt_str = batt_str
            # sleep if nothing has occurred:
            if not send_event and not recv_event:
                time.sleep(0.1)
        # close the communication channels:
        send_q.close()
        recv_q.close()

    def __listener_thread(self):
        """
        Thread in charge of receiving UDP messages.
        """
        while True:
            mes = self.udp_int.recv()
            for mb in mes:
                m = pkl.loads(mb)
                if m[0] == "ACK":  # aknowledgement type
                    self._lock.acquire()
                    if m[1] > self.__last_received_id:
                        self.__last_received_id = m[1]  # identifier
                    self._lock.release()
                elif m[0] == "OBS":  # observation type
                    self._lock.acquire()
                    i_obs = m[1]
                    if i_obs > self.__last_i_obs:  # drop outdated observations
                        self.__obs = m[2]  # observation
                        self.__last_i_obs = i_obs
                    self._lock.release()
                elif m[0] == "RES":  # acknowledgement for the reset command
                    self._lock.acquire()
                    self.__wait_ack_reset = False
                    self._lock.release()
                elif m[0] == "DON":  # 'done' signal from position control
                    self._lock.acquire()
                    if self.__last_sent_id == m[1]:
                        self.__wait_done = False
                    self._lock.release()
                elif m[0] == "BBT":  # bad battery state
                    with self._gui_lock:
                        self._gui_batt_str = f"{m[1][0]}V"

    def __resender_thread(self):
        """
        Thread in charge of resending the last UDP message when it has not been acknowledged by the drone.
        """
        while True:
            time.sleep(self.wait_ack_duration)
            self._lock.acquire()
            if self.__last_received_id < self.__last_sent_id:
                t = time.time()
                if t - self.__last_timestamp > self.wait_ack_duration:
                    self.udp_int.send(self.__last_sent)  # resend (with the same identifier as before)
                    self.__last_timestamp = t
            self._lock.release()

    def send(self, msg_type, msg=None):
        """
        Send an UDP message to the drone.
        Args:
            msg_type: string in ["RES", ACT"].
            msg: tuple of the form (str:command, ...:args).
        """
        self._lock.acquire()
        self.__last_sent_id += 1
        self.__last_timestamp = time.time()
        data = (msg_type, self.__last_sent_id, msg) if msg is not None else (msg_type, self.__last_sent_id)
        data_s = pkl.dumps(data)
        self.__last_sent = data_s
        self.udp_int.send(data_s)
        self._lock.release()

    def sleep_until_done(self, max_duration, granularity=0.05):
        """
        Sleeps until either max_duration is elapsed or a 'done" signal is received from the position control API.
        """
        timeout_t = time.time() + max_duration
        while True:
            self._lock.acquire()
            c = self.__wait_done
            self._lock.release()
            t = time.time()
            if not c or t >= timeout_t:
                if c:
                    self._lock.acquire()
                    self.__wait_done = False
                    self._lock.release()
                break
            time.sleep(granularity)

    def wait_acknowledgement(self, max_duration=1.0, granularity=0.05):
        """
        sleeps until an acknowledgement is received from the drone or max_duration is elapsed
        """
        timeout_t = time.time() + max_duration
        while True:
            self._lock.acquire()
            c = self.__last_received_id < self.__last_sent_id
            self._lock.release()
            t = time.time()
            if not c or t >= timeout_t:
                break
            time.sleep(granularity)

    # ==================================================================================================================

    # Pro API
    # Use this except for school purpose:

    def reset(self, sleep_duration=0.1):
        """
        Resets the drone and waits for reset acknowlegement.
        Args:
            sleep_duration: float: the reset command waits for at least this amount of time
        """
        self.send(msg_type="RES", msg=(self.local_ip, self.recv_port))
        # wait for the RES message to return
        c = True
        logger.info("Waiting for reset acknowlegement...")
        while c:
            time.sleep(sleep_duration)
            self._lock.acquire()
            c = self.__wait_ack_reset
            if not self.__wait_ack_reset:
                self.__wait_ack_reset = True  # for next reset
            self._lock.release()
        logger.info("Reset acknowlegement received.")
        self._lock.acquire()
        self.__obs = None
        self.__last_sent_id = 0
        self.__last_received_id = 0
        self.__last_timestamp = time.time()
        self.__last_sent = None
        self._lock.release()
        if self.wait_for_first_obs:
            logger.info("waiting for a first observation from the drone")
            while True:
                time.sleep(self.wait_for_first_obs_sleep_duration)
                self._lock.acquire()
                obs = deepcopy(self.__obs)
                self._lock.release()
                if obs is not None:
                    break
            logger.info("observation initialized")

    def arm(self):
        """
        Arms the drone.
        """
        self.send(msg_type="ACT", msg=("ARM",))

    def disarm(self):
        """
        Disarms the drone.
        """
        self.send(msg_type="ACT", msg=("DISARM",))

    def takeoff_nonblocking(self, altitude=None):
        """
        The drone takes off.

        Do not use altitude (this currently refers to a FC throttle value and is likely to change in the future).
        """
        self.send(msg_type="ACT", msg=("TAKEOFF", altitude))
        self.time_takeoff = time.time()

    def land_nonblocking(self):
        """
        The drone lands.
        """
        self.send(msg_type="ACT", msg=("LAND",))

    def set_velocity_nonblocking(self, v_x=0.0, v_y=0.0, v_z=0.0, w=0.0, duration=1.0, drone_frame=True):
        """
        Sets the drone velocity.
        The drone needs to be armed.
        Args:
            v_x: float (optional): velocity target on the frontward axis (m/s); tested range: 0.1 - 0.5
            v_y: float (optional): velocity target on the rightward axis (m/s); tested range: 0.1 - 0.5
            v_z: float (optional): velocity target on the upward axis (m/s); tested range: 0.05 - 0.1
            w: float (optional): yaw rate on the upward axis (rad/s)); tested range: 0.05 - 0.5
            duration: float (optional): maximum max_duration of the command (s)
            drone_frame: bool (optional): whether to use the drone frame (True) or the world frame (False)
        """
        frame_str = "VDF" if drone_frame else "VWF"
        self.send(msg_type="ACT", msg=(frame_str, v_x, v_y, v_z, w, duration))

    def set_position_nonblocking(self, x, y, z, yaw=None, max_velocity=0.5, max_yaw_rate=0.5, max_duration=10.0, relative=False):
        """
        Sets a position target for the drone.
        If relative is False (default), the target is relative to the world axis.
        If relative is True, the target is relative to the drone frame.
        Note: z is always in the world frame.
        The drone needs to be armed.
        Args:
            x: float: x target (m)
            y: float: y target (m)
            z: float: z target (m); tested range: 0.3 - 0.9
            yaw: float (optional): yaw target (rad); if None, the yaw is not changed
            max_velocity: float (optional): maximum velocity used to go to position (m/s)
            max_yaw_rate: float (optional): maximum yaw rate used to go to yaw (rad/s)
            max_duration: float (optional): maximum duration of the command (s)
            relative: bool (optional): whether to use the drone frame (True) or the world frame (False, default)
        """
        if relative:
            frame_str = "PDF"
        else:
            frame_str = "PWF"
        self._lock.acquire()
        self.__wait_done = True
        self._lock.release()
        self.send(msg_type="ACT", msg=(frame_str, x, y, z, yaw, max_velocity, max_yaw_rate, max_duration))

    def set_pid_vel_x(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
        """
        Sets the gains of the PID for the x velocity tracker.
        Gains left as None are not affected.
        Example of reasonable values:
            k_p = 750.0
            k_i = 400.0
            k_d = 50.0
        Args:
            k_p: float (optional): proportional gain
            k_i: float (optional): integral gain
            k_d: float (optional): derivative gain
            timeout: float (optional): timeout when waiting for confirmation from the drone
        """
        self.send(msg_type="PVX", msg=(k_p, k_i, k_d))
        self.wait_acknowledgement(max_duration=timeout)

    def set_pid_vel_y(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
        """
        Sets the gains of the PID for the y velocity tracker.
        Gains left as None are not affected.
        Example of reasonable values:
            k_p = 750.0
            k_i = 400.0
            k_d = 50.0
        Args:
            k_p: float (optional): proportional gain
            k_i: float (optional): integral gain
            k_d: float (optional): derivative gain
            timeout: float (optional): timeout when waiting for confirmation from the drone
        """
        self.send(msg_type="PVY", msg=(k_p, k_i, k_d))
        self.wait_acknowledgement(max_duration=timeout)

    def set_pid_vel_z(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
        """
        Sets the gains of the PID for the z velocity tracker.
        Gains left as None are not affected.
        Example of reasonable values:
            k_p = 50.0
            k_i = 20.0
            k_d = 5.0
        Args:
            k_p: float (optional): proportional gain
            k_i: float (optional): integral gain
            k_d: float (optional): derivative gain
            timeout: float (optional): timeout when waiting for confirmation from the drone
        """
        self.send(msg_type="PVZ", msg=(k_p, k_i, k_d))
        self.wait_acknowledgement(max_duration=timeout)

    def set_pid_vel_w(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
        """
        Sets the gains of the PID for the yaw rate tracker.
        Gains left as None are not affected.
        Example of reasonable values:
            k_p = 400.0
            k_i = 200.0
            k_d = 0.0
        Args:
            k_p: float (optional): proportional gain
            k_i: float (optional): integral gain
            k_d: float (optional): derivative gain
            timeout: float (optional): timeout when waiting for confirmation from the drone
        """
        self.send(msg_type="PVW", msg=(k_p, k_i, k_d))
        self.wait_acknowledgement(max_duration=timeout)

    # ==================================================================================================================

    # High-level API for school purpose
    # (don't use along the pro API, otherwise weird things will happen with altitude)
    # (sleeps after calls, uses cm instead of m, and uses degrees instead of rad):

    def takeoff(self, altitude=None, alt_duration=10.0):
        """
        Arms the drone and takes off
        Args:
            altitude: float (optional): target altitude (cm)
            alt_duration: float (optional): additional max duration to reach the target altitude
        """
        altitude = altitude / 100.0 if altitude is not None else EASY_API_TAKEOFF_ALTITUDE
        self.reset()
        self.arm()
        time.sleep(1.0)
        # the takeoff altitude might depend on the battery with this and Z will only be defined properly later:
        self.takeoff_nonblocking()
        time.sleep(10.0)
        self.easy_api_cur_z = altitude
        # this works (Z properly defined) but when done from the ground it is a bit violent and not very stable:
        self.set_position_nonblocking(x=0.0, y=0.0, z=self.easy_api_cur_z, yaw=0.0,
                                      max_velocity=self.ea_speed,
                                      max_yaw_rate=self.ea_yaw_rate,
                                      max_duration=alt_duration,
                                      relative=False)
        self.sleep_until_done(alt_duration)

    def land(self, sleep_duration=5.0):
        """
        Lands and disarms the drone
        """
        self.easy_api_cur_z = 0.0
        self.land_nonblocking()
        time.sleep(sleep_duration)
        self.disarm()

    def forward(self, val_cm):
        """
        Goes forward by number of centimeters
        """
        val_m = val_cm / 100.0
        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
        self.set_position_nonblocking(x=val_m, y=0.0, z=self.easy_api_cur_z, yaw=None,
                                      max_velocity=self.ea_speed,
                                      max_yaw_rate=self.ea_yaw_rate,
                                      max_duration=duration,
                                      relative=True)
        self.sleep_until_done(duration)

    def backward(self, val_cm):
        """
        Goes backward by number of centimeters
        """
        val_m = val_cm / 100.0
        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
        self.set_position_nonblocking(x=-val_m, y=0.0, z=self.easy_api_cur_z, yaw=None,
                                      max_velocity=self.ea_speed,
                                      max_yaw_rate=self.ea_yaw_rate,
                                      max_duration=duration,
                                      relative=True)
        self.sleep_until_done(duration)

    def right(self, val_cm):
        """
        Goes right by number of centimeters
        """
        val_m = val_cm / 100.0
        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
        self.set_position_nonblocking(x=0.0, y=val_m, z=self.easy_api_cur_z, yaw=None,
                                      max_velocity=self.ea_speed,
                                      max_yaw_rate=self.ea_yaw_rate,
                                      max_duration=duration,
                                      relative=True)
        self.sleep_until_done(duration)

    def left(self, val_cm):
        """
        Goes left by number of centimeters
        """
        val_m = val_cm / 100.0
        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
        self.set_position_nonblocking(x=0.0, y=-val_m, z=self.easy_api_cur_z, yaw=None,
                                      max_velocity=self.ea_speed,
                                      max_yaw_rate=self.ea_yaw_rate,
                                      max_duration=duration,
                                      relative=True)
        self.sleep_until_done(duration)

    def up(self, val_cm):
        """
        Goes up by number of centimeters
        """
        val_m = val_cm / 100.0
        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
        self.easy_api_cur_z += val_m
        self.easy_api_cur_z = clip(self.easy_api_cur_z, EASY_API_MIN_ALTITUDE, EASY_API_MAX_ALTITUDE)
        self.set_position_nonblocking(x=0.0, y=0.0, z=self.easy_api_cur_z, yaw=None,
                                      max_velocity=self.ea_speed,
                                      max_yaw_rate=self.ea_yaw_rate,
                                      max_duration=duration,
                                      relative=True)
        self.sleep_until_done(duration)

    def down(self, val_cm):
        """
        Goes down by number of centimeters
        """
        val_m = val_cm / 100.0
        duration = val_m / self.ea_speed + EASY_API_ADDITIONAL_DURATION
        self.easy_api_cur_z -= val_m
        self.easy_api_cur_z = clip(self.easy_api_cur_z, EASY_API_MIN_ALTITUDE, EASY_API_MAX_ALTITUDE)
        self.set_position_nonblocking(x=0.0, y=0.0, z=self.easy_api_cur_z, yaw=None,
                                      max_velocity=self.ea_speed,
                                      max_yaw_rate=self.ea_yaw_rate,
                                      max_duration=duration,
                                      relative=True)
        self.sleep_until_done(duration)

    def cw(self, val_deg):
        """
        Rotates clockwise by number of degrees (range 0 - 180).
        Note: the drone takes the shortest path, direction is not guaranteed for angles close to 180.
        """
        # val_rad = max(val_deg * np.pi / 180.0, 0.0)
        # duration = val_rad / self.ea_yaw_rate + EASY_API_ADDITIONAL_DURATION
        # seq_angles = get_angle_sequence_rad(val_rad)
        # seq = []
        # for sa in seq_angles:
        #     seq.append([0, 0, self.easy_api_cur_z, sa])
        # self.position_sequence(sequence=seq, max_duration=duration, relative=True)

        val_rad = np.pi if val_deg == 180 else (val_deg % 180) * np.pi / 180.0
        duration = val_rad / self.ea_yaw_rate + EASY_API_ADDITIONAL_DURATION
        self.set_position_nonblocking(x=0.0, y=0.0, z=self.easy_api_cur_z, yaw=val_rad,
                                      max_velocity=self.ea_speed,
                                      max_yaw_rate=self.ea_yaw_rate,
                                      max_duration=duration,
                                      relative=True)
        self.sleep_until_done(duration)

    def ccw(self, val_deg):
        """
        Rotates counter-clockwise by number of degrees (range 0 - 180).
        Note: the drone takes the shortest path, direction is not guaranteed for angles close to 180.
        """
        # val_rad = max(val_deg * np.pi / 180.0, 0.0)
        # duration = val_rad / self.ea_yaw_rate + EASY_API_ADDITIONAL_DURATION
        # val_rad = - val_rad
        # seq_angles = get_angle_sequence_rad(val_rad)
        # seq = []
        # for sa in seq_angles:
        #     seq.append([0, 0, self.easy_api_cur_z, sa])
        # self.position_sequence(sequence=seq, max_duration=duration, relative=True)

        val_rad = np.pi if val_deg == 180 else (val_deg % 180) * np.pi / 180.0
        duration = val_rad / self.ea_yaw_rate + EASY_API_ADDITIONAL_DURATION
        self.set_position_nonblocking(x=0.0, y=0.0, z=self.easy_api_cur_z, yaw=-val_rad,
                                      max_velocity=self.ea_speed,
                                      max_yaw_rate=self.ea_yaw_rate,
                                      max_duration=duration,
                                      relative=True)
        self.sleep_until_done(duration)

    def go(self, x, y, z, yaw=None, max_duration=10.0):
        """
        Goes to the requested position (cm) and yaw (deg) targets.
        If yaw is None, no yaw target is set.
        """
        x = x / 100.0
        y = y / 100.0
        z = z / 100.0
        y = y * np.pi / 180.0
        self.easy_api_cur_z = clip(z, EASY_API_MIN_ALTITUDE, EASY_API_MAX_ALTITUDE)
        self.set_position_nonblocking(x=x, y=y, z=self.easy_api_cur_z, yaw=yaw,
                                      max_velocity=self.ea_speed,
                                      max_yaw_rate=self.ea_yaw_rate,
                                      max_duration=max_duration,
                                      relative=False)
        self.sleep_until_done(max_duration)

    def position_sequence(self, sequence, max_duration=60.0, relative=False, speed=None, yaw_rate=None):
        """
        The drone follows a roadmap defined by a sequence of targets.
        Args:
            sequence: a sequence of sequence-like elements, each of length 3 or 4 (mixing 3 and 4 is possible).
                If the length of an element is 3, it is interpreted as (x, y, z) (cm).
                If the length of an element is 4, it is interpreted as (x, y, z, yaw) (cm and deg).
            max_duration: float: maximum duration of the whole command.
            relative: bool: whether x, y and yaw have to be interpreted in the drone frame (True)
                or in the world frame (False, default).
                Note: z is always in the world frame.
            speed: float: if not None, target speed in cm/s, else the default is used.
            yaw_rate: float: if not None, target yaw rate in deg/s, else the default is used.
        """
        if speed is not None:
            speed = speed / 100.0  # convert to m/s
            speed = max(speed, 0.0)
        if yaw_rate is not None:
            yaw_rate = yaw_rate * np.pi / 180.0  # convert to rad/s
            yaw_rate = max(yaw_rate, 0.0)
        remaining_duration = max_duration
        t_start = time.time()
        for elt in sequence:
            if 3 <= len(elt) <= 4 and remaining_duration > 0:
                x = elt[0] / 100.0  # convert to cm
                y = elt[1] / 100.0  # convert to cm
                z = elt[2] / 100.0  # convert to cm
                yaw = elt[3] * np.pi / 180.0 if len(elt) == 4 else None  # convert to rad

                self.easy_api_cur_z = clip(z, EASY_API_MIN_ALTITUDE, EASY_API_MAX_ALTITUDE)
                self.set_position_nonblocking(x=x, y=y, z=self.easy_api_cur_z, yaw=yaw,
                                              max_velocity=self.ea_speed if speed is None else speed,
                                              max_yaw_rate=self.ea_yaw_rate if yaw_rate is None else yaw_rate,
                                              max_duration=remaining_duration,
                                              relative=relative)
                self.sleep_until_done(remaining_duration)
                elapsed_time = time.time() - t_start
                remaining_duration -= elapsed_time

    def curve(self, x1, y1, z1, x2, y2, z2, speed):
        """
        2-positions roadmap.
        This is an alias for position_sequence(sequence=[[x1, y1, z1],[x2, y2, z2]], speed=speed).
        """
        self.position_sequence(sequence=[[x1, y1, z1],[x2, y2, z2]], speed=speed)

    def set_speed(self, speed):
        """
        Sets the maximum speed of the drone in the school API
        Args:
            speed: float: speed in cm/s; a reasonable value is 20.
        """
        self.ea_speed = speed / 100.0

    def set_yaw_rate(self, yaw_rate):
        """
        Sets the maximum yaw rate of the drone in the school API
        Args:
            yaw_rate: float: yaw rate in deg/s; a reasonable value is 30.
        """
        self.ea_yaw_rate = yaw_rate * np.pi / 180.0

    # ==================================================================================================================

    # Methods common to both APIs:

    def streamon(self, resolution="VGA", fps=10, display=False, wait_first_frame=True):
        """
        Starts camera streaming.
        CAUTION: This will slow the frequency of the onboard controller down and may make the drone unstable!
        The highest resolution and fps are, the worst the influence on the onboard controller.
        The image transfers happens over TCP, which comes with a noticeable delay.
        CAUTION: High fps may saturate the network in the presence of several drones.
        Args:
            resolution: Tuple(float: height, float: width): resolution of the captured images.
            fps: integer: maximum framerate (may not be attained),
            display: boolean: whether to display the stream in the GUI (requires GUI to be enabled),
            wait_first_frame: boolean: whether to sleep until the first frame is available.
        """
        if not self.tcp_video_int.is_receiver_running():
            free_port = get_free_port(self.recv_port_video)
            assert free_port is not None, "No available port!"
            if free_port != self.recv_port_video:
                logger.warning(f"Port {self.recv_port_video} is unavailable, trying to communicate on port {free_port} instead.")
                self.recv_port_video = free_port
            self.tcp_video_int.start_receiver(self.recv_port_video)
            time.sleep(1.0)  # sleep a bit so the server starts before the drone tries to connect
            self.send(msg_type="ST1", msg=(self.local_ip, self.recv_port_video, resolution, fps))
            if wait_first_frame:
                _ = self.get_frame(wait_new_frame=True)
            self._gui_lock.acquire()
            self._gui_display = display
            self._gui_lock.release()

    def stream(self, resolution="VGA", fps=10):
        """
        Alias for streamon(resolution, fps=fps, display=True)
        """
        self.streamon(resolution, fps, display=True)

    def get_frame(self, wait_new_frame=True, timeout=10.0, sleep_between_trials=0.05):
        """
        Retrieves a single frame from the stream.
        Args:
            wait_new_frame: boolean: if True, two consecutive calls to the fonction will not return identical frames.
            timeout: after this duration without retrieving a frame, a TimeoutException will be raised.
            sleep_between_trials: float: will sleep this amount of time between two consecutive trials or retrieval.
        """
        t_start = time.time()
        im = None
        im_id = self.last_im_id if wait_new_frame else -1
        cond = True
        while cond:
            im, im_id = self.tcp_video_int.get_last_image()
            cond1 = im is None and im_id <= self.last_im_id
            cond = cond1 and time.time() - t_start < timeout
            if cond:
                time.sleep(sleep_between_trials)
            else:
                self.last_im_id = im_id
        if cond1:
            raise TimeoutError("Could not retrieve frame from stream.")
        return im

    def streamoff(self):
        """
        Stops camera streaming
        """
        self._gui_lock.acquire()
        self._gui_display = False
        self._gui_lock.release()
        self.send(msg_type="ST0")
        time.sleep(1.0)
        self.tcp_video_int.stop_receiver()

    def get_time(self):
        """
        Gets the number of seconds since the last takeoff command was sent
        """
        return time.time() - self.time_takeoff

    def get_telemetry(self):
        """
        Gets a tuple that describe the state of the drone in the world frame.
        Angles are in radians.
        Returns:
            telemetry: Tuple: (battery_voltage: float,
                               x: float,
                               y: float,
                               z: float,
                               yaw: float,
                               vel_x: float,
                               vel_y: float,
                               vel_z: float,
                               yaw_rate: float,
                               health_flags: list of strings)
        """
        self._lock.acquire()
        obs = deepcopy(self.__obs)
        self._lock.release()
        return obs

    def get_battery(self):
        """
        Returns:
            battery_voltage: float: battery voltage, in V
        """
        telemetry = self.get_telemetry()
        return telemetry[0]

    def get_position(self):
        """
        Returns:
            x: float: in cm
            y: float: in cm
            z: float: in cm
        """
        telemetry = self.get_telemetry()
        return telemetry[1] * 100, telemetry[2] * 100, telemetry[3] * 100

    def get_yaw(self):
        """
        Returns:
            x: float: yaw, in deg
        """
        telemetry = self.get_telemetry()
        return telemetry[4] * 180.0 / np.pi

    def get_velocity(self):
        """
        Returns:
            v_x: float: in cm/s
            v_y: float: in cm/s
            v_z: float: in cm/s
        """
        telemetry = self.get_telemetry()
        return telemetry[5] * 100, telemetry[6] * 100, telemetry[7] * 100

    def get_health(self):
        """
        Returns:
            health_flags: list os strings: useful flags for debugging
        """
        telemetry = self.get_telemetry()
        return telemetry[9]

    def get_yaw_rate(self):
        """
        Returns:
            w: float: yaw rate in deg/s
        """
        telemetry = self.get_telemetry()
        return telemetry[8] * 180.0 / np.pi

    def get_tof(self):
        """
        Returns:
            h: float: altitude in cm
        """
        _, _, res = self.get_position()
        return res

    def get_height(self):
        """
        Alias for get_tof() as cognifly doesn't know its height from the starting point.
        """
        return self.get_tof()

    def get_speed(self):
        """
        Returns:
            speed: float: norm of the velocity, in cm/s
        """
        v_x, v_y, v_z = self.get_velocity()
        return np.linalg.norm([v_x, v_y, v_z])



import socket
from psutil import process_iter
import time
import warnings


def extract_ip():
    st = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        st.connect(('10.255.255.255', 1))
        ip = st.getsockname()[0]
    except Exception:
        ip = '127.0.0.1'
    finally:
        st.close()
    return ip


def get_free_port(port):
    initial_port = port
    is_free = True
    while True:
        for proc in process_iter():
            try:
                for conns in proc.connections(kind='inet'):
                    if conns.laddr.port == port:
                        is_free = False
            except:
                pass
        if is_free:
            break
        else:
            is_free = True
            port += 1
            if port >= 65000:
                port = 2000
            if port == initial_port:
                return None
    return port



import time
from collections import deque
from itertools import cycle
import socket
import pickle as pkl
import numpy as np
from pathlib import Path

from cognifly.utils.udp_interface import UDPInterface
from cognifly.utils.ip_tools import extract_ip
from cognifly.utils.pid import PID
from cognifly.utils.filters import Simple1DKalman, Simple1DExponentialAverage
from cognifly.utils.functions import clip, smallest_angle_diff_rad


KEY_TIMEOUT = 0.2  # seconds before a keyboard command times out

DEFAULT_ROLL = 1500
DEFAULT_PITCH = 1500
DEFAULT_THROTTLE = 900  # throttle bellow a certain value disarms the FC
DEFAULT_YAW = 1500

TAKEOFF = 1300
LAND = 900

MIN_CMD_ROLL = 1000
MIN_CMD_PITCH = 1000
MIN_CMD_THROTTLE = 900  # throttle bellow a certain value disarms the FC
MIN_CMD_YAW = 1400
MAX_CMD_ROLL = 2000
MAX_CMD_PITCH = 2000
MAX_CMD_THROTTLE = 1500
MAX_CMD_YAW = 1600

KEY_N_ROLL = 1350
KEY_N_PITCH = 1350
KEY_N_THROTTLE = -10  # added on throttle(-) key press
KEY_N_YAW = 1450
KEY_P_ROLL = 1650
KEY_P_PITCH = 1650
KEY_P_THROTTLE = 10  # added on throttle(+) key press
KEY_P_YAW = 1550

KEY_TAKEOFF = TAKEOFF
KEY_LAND = LAND

# DEFAULT_CH5 = 1000  # DISARMED (1000) / ARMED (1800)
# DEFAULT_CH6 = 1500  # ANGLE (1000) / NAV_POSHOLD (1500) <= 1300: NAV ALTHOLD
# DEFAULT_CH7 = 1000  # FAILSAFE (1800)
# DEFAULT_CH8 = 1000  # HEADING HOLD (1800)

ARMED = 1800
DISARMED = 1000
ANGLE_MODE = 1000
NAV_ALTHOLD_MODE = 1500
NAV_POSHOLD_MODE = 1800

DEFAULT_AUX1 = DISARMED  # DISARMED (1000) / ARMED (1800)
DEFAULT_AUX2 = NAV_POSHOLD_MODE  # Angle (1000) / NAV_ALTHOLD (1500) / NAV_POSHOLD (1800)

# Max periods for:
CTRL_LOOP_TIME = 1 / 100
SLOW_MSGS_LOOP_TIME = 1 / 5  # these messages take a lot of time slowing down the loop...
BATT_UDP_MSG_TIME = 1.0

NO_OF_CYCLES_AVERAGE_GUI_TIME = 10

BATT_NO_INFO = -1
BATT_OK = 0
BATT_TOO_HIGH = 1
BATT_WARNING = 2
BATT_TOO_LOW = 3

EPSILON_DIST_TO_TARGET = 0.05  # (m)
EPSILON_ANGLE_TO_TARGET = 1 * np.pi / 180.0  # (rad)


class SignalTracer:
    """
    This can be used to trace the drone dynamics in real-time (e.g. to output a CSV-like log file)
    """
    def __init__(self, filepath):
        self.file = open(filepath, 'w+')

    def write_line(self, line):
        self.file.write(line + '\n')

    def __del__(self):
        self.file.close()


class CogniflyControllerDummy:
    def __init__(self,
                 network=True,
                 drone_hostname=None,
                 drone_port=8988,
                 print_screen=True,
                 obs_loop_time=0.1,
                 trace_logs=False):
        self.network = network
        if not self.network:
            self.udp_int = None
            self.drone_hostname = None
            self.drone_port = None
        else:
            self.udp_int = UDPInterface()
            self.drone_hostname = socket.gethostname() if drone_hostname is None else drone_hostname
            self.drone_ip = socket.gethostbyname(self.drone_hostname) if drone_hostname is not None else extract_ip()
            self.drone_port = drone_port
            self.udp_int.init_receiver(ip=self.drone_ip, port=self.drone_port)
            print(f"receiver initialized with ip:{self.drone_ip} and port {self.drone_port}")
        self.sender_initialized = False  # sender is initialized only when a reset message is received
        self.print_screen = print_screen
        self.obs_loop_time = obs_loop_time
        self.voltage = None
        self.min_voltage = None
        self.warn_voltage = None
        self.max_voltage = None
        self.key_cmd_in_progress = False
        self.udp_armed = False
        self.ludp = 0
        self.last_key_tick = time.time()
        self.last_obs_tick = time.time()
        self.last_bat_tick = time.time()
        # reminder:
        # 'CH5': DEFAULT_CH5  # DISARMED (1000) / ARMED (1800)
        # 'CH6': DEFAULT_CH6  # ANGLE (1000) / NAV_POSHOLD (1500)
        # 'CH7': DEFAULT_CH7  # FAILSAFE (1800)
        # 'CH8': DEFAULT_CH8  # HEADING HOLD (1800)
        self.CMDS = {'roll': DEFAULT_ROLL,
                     'pitch': DEFAULT_PITCH,
                     'throttle': DEFAULT_THROTTLE,  # throttle bellow a certain value disarms the FC
                     'yaw': DEFAULT_YAW,
                     'aux1': DEFAULT_AUX1,
                     'aux2': DEFAULT_AUX2
                     }
        self.CMDS_ORDER = ['roll', 'pitch', 'throttle', 'yaw', 'aux1', 'aux2']
        self.__batt_state = BATT_NO_INFO

        self.current_flight_command = None
        self._flight_origin = None
        self.prev_yaw_prime_ts = None
        self.prev_alt_prime_ts = None

        self.telemetry = None
        self.debug_flags = []
        self._i_obs = 0
        self.target_flag = False
        self.target_id = None

        # PID values:

        self.pid_limits_xy = (-500, +500)
        self.pid_limits_z = (-500, +500)
        self.pid_limits_w = (-500, +500)

        self.vel_x_kp = 750.0
        self.vel_x_ki = 400.0
        self.vel_x_kd = 50.0

        self.vel_y_kp = 750.0
        self.vel_y_ki = 400.0
        self.vel_y_kd = 50.0

        self.vel_z_kp = 50.0
        self.vel_z_ki = 20.0
        self.vel_z_kd = 5.0

        self.vel_w_kp = 400.0  # 500
        self.vel_w_ki = 200.0  # 100
        self.vel_w_kd = 0.0

        self.pid_vel_x = None
        self.pid_vel_y = None
        self.pid_vel_z = None
        self.pid_w_z = None

        self.x_vel_gain = 1
        self.y_vel_gain = 1
        self.z_vel_gain = 1
        self.w_gain = 0.5

        # filters:

        # self.filter_vel_z = Simple1DKalman(err_measure=0.3, q=0.9)
        # self.filter_w_z = Simple1DKalman(err_measure=0.2, q=0.9)
        self.filter_vel_z = Simple1DExponentialAverage(tau=0.2)
        self.filter_w_z = Simple1DExponentialAverage(tau=0.2)

        # tracer:

        self.start_time = time.time()
        self.trace_logs = trace_logs
        if self.trace_logs:
            self.w_tracer = SignalTracer(Path(__file__).parent.resolve() / "trace_logs" / "w_trace.log")
            self.z_tracer = SignalTracer(Path(__file__).parent.resolve() / "trace_logs" / "z_trace.log")
            self.x_tracer = SignalTracer(Path(__file__).parent.resolve() / "trace_logs" / "x_trace.log")
            self.y_tracer = SignalTracer(Path(__file__).parent.resolve() / "trace_logs" / "y_trace.log")
        else:
            self.w_tracer = None
            self.z_tracer = None
            self.x_tracer = None
            self.y_tracer = None

    def run_curses(self):
        result = 1
        try:
            result = self._controller()
        finally:
            if result == 1:
                print("An error occurred, probably the serial port is not available")

    def _udp_commands_handler(self, command):
        """
        Args:
            command: a command of the form (str:command, ...:args)
        """
        print(f"handling command {command}")
        if command[0] == "RES":  # reset drone
            self.CMDS = {'roll': DEFAULT_ROLL,
                         'pitch': DEFAULT_PITCH,
                         'throttle': DEFAULT_THROTTLE,  # throttle bellow a certain value disarms the FC
                         'yaw': DEFAULT_YAW,
                         'aux1': DISARMED,  # disarmed
                         'aux2': DEFAULT_AUX2
                         }
            self.udp_int.init_sender(ip=command[2][0], port=command[2][1])
            self.sender_initialized = True
            self._flight_origin = None
            self.udp_int.send(pkl.dumps(("RES", (self.drone_ip, self.drone_port))))
            self.current_flight_command = None
        elif self.sender_initialized:  # do not take any udp order unless reset has been called
            identifier = command[1]
            if command[0] == "ACT":  # action
                # hard-reset all PIDs (FIXME: this is to avoid a bug in the PID framework that seems to retain unsuitable values otherwise, correct this in the future)
                self.pid_vel_x = PID(kp=self.vel_x_kp, ki=self.vel_x_ki, kd=self.vel_x_kd, sample_time=0.01, output_limits=self.pid_limits_xy, auto_mode=False)
                self.pid_vel_y = PID(kp=self.vel_y_kp, ki=self.vel_y_ki, kd=self.vel_y_kd, sample_time=0.01, output_limits=self.pid_limits_xy, auto_mode=False)
                self.pid_vel_z = PID(kp=self.vel_z_kp, ki=self.vel_z_ki, kd=self.vel_z_kd, sample_time=0.01, output_limits=self.pid_limits_z, auto_mode=False)
                self.pid_w_z = PID(kp=self.vel_w_kp, ki=self.vel_w_ki, kd=self.vel_w_kd, sample_time=0.01, output_limits=self.pid_limits_w, auto_mode=False)
                if command[2][0] == "DISARM":
                    self.CMDS["aux1"] = DISARMED
                    self.current_flight_command = None
                elif command[2][0] == "ARM":
                    self.CMDS["aux1"] = ARMED
                    self.CMDS["aux2"] = NAV_POSHOLD_MODE
                    self.current_flight_command = None
                elif command[2][0] == "TAKEOFF":
                    alt = command[2][1] if command[2][1] is not None else TAKEOFF
                    self.CMDS["throttle"] = alt  # TODO: replace by true altitude (now a throttle value)
                    self.CMDS["aux2"] = NAV_POSHOLD_MODE
                    self.current_flight_command = None
                elif command[2][0] == "LAND":
                    self.CMDS["throttle"] = LAND
                    self.CMDS["aux2"] = NAV_POSHOLD_MODE
                    self.current_flight_command = None
                elif command[2][0] in ("VDF", "VWF"):  # velocity
                    self.current_flight_command = [command[2][0], command[2][1], command[2][2], command[2][3], command[2][4], time.time() + command[2][5]]
                elif command[2][0] in ("PDF", "PWF"):  # position
                    self.target_flag = True
                    self.target_id = identifier
                    self.current_flight_command = [command[2][0], command[2][1], command[2][2], command[2][3], command[2][4], command[2][5], command[2][6], time.time() + command[2][7]]
            elif command[0] == "ST1":  # stream on
                pass
            elif command[0] == "ST0":  # stream off
                pass
            self.udp_int.send(pkl.dumps(("ACK", identifier)))

    def _compute_z_vel(self, pos_z_wf):
        if self.prev_alt_prime_ts is None:
            self.prev_alt_prime_ts = (pos_z_wf, 0.0, time.time())
            res = 0.0
        else:
            now = time.time()
            dt = now - self.prev_alt_prime_ts[2]
            if dt <= 0:
                res = self.prev_alt_prime_ts[1]
            else:
                res = (pos_z_wf - self.prev_alt_prime_ts[0]) / dt
                self.prev_alt_prime_ts = (pos_z_wf, res, now)
        before_filter = res
        res = self.filter_vel_z.update_estimate(res)
        if self.trace_logs:
            self.z_tracer.write_line(f"{self.prev_alt_prime_ts[2]-self.start_time};{self.prev_alt_prime_ts[0]};{before_filter};{res}")
        return res

    def _compute_yaw_rate(self, yaw):
        if self.prev_yaw_prime_ts is None:
            self.prev_yaw_prime_ts = (yaw, 0.0, time.time())
            res = 0.0
        else:
            now = time.time()
            dt = now - self.prev_yaw_prime_ts[2]
            if dt <= 0:
                res = self.prev_yaw_prime_ts[1]
            else:
                # to handle jumps in yaw, we take the smallest abolute value
                diff = smallest_angle_diff_rad(yaw, self.prev_yaw_prime_ts[0])
                res = diff / dt
                self.prev_yaw_prime_ts = (yaw, res, now)
        before_filter = res
        res = self.filter_w_z.update_estimate(res)
        if self.trace_logs:
            self.w_tracer.write_line(f"{self.prev_yaw_prime_ts[2]-self.start_time};{self.prev_yaw_prime_ts[0]};{before_filter};{res}")
        return res

    def _flight(self):
        """
        This method is a high-level flight controller.
        Depending on the current flight command and state of the drone, it outputs relevant MSP commands.
        This is done thanks to python PIDs and filters.
        In the future, most of this should be integrated directly in the inav flight controller.
        Indeed, the use of PIDs here is redundant with what happens in inav.
        However, the cognifly version of inav will not take position/velocities as input at the moment but only RPYT, hence this method.
        The use of PIDs for velocities should be removed fairly easily, if we can find the mapping from velocities to RPYT.
            This is the inverse of a monotonic polynomial for x and y, see inav code (potentially no closed form?).
            We have not found the equivalent for z and yaw in inav code yet.
        """
        # retrieve the current state
        yaw = 0.0
        yaw_rate = self._compute_yaw_rate(yaw)
        pos_x_wf = 0.0
        pos_y_wf = 0.0
        vel_x_wf = 0.0
        vel_y_wf = 0.0
        pos_z_wf = 0.0
        vel_z_wf = self._compute_z_vel(pos_z_wf)

        if self._flight_origin is None:  # new basis at reset
            self._flight_origin = (pos_x_wf, pos_y_wf, yaw)
        else:
            # change of basis
            x_int = pos_x_wf - self._flight_origin[0]
            y_int = pos_y_wf - self._flight_origin[1]
            cos = np.cos(self._flight_origin[2])
            sin = np.sin(self._flight_origin[2])
            pos_x_wf = x_int * cos + y_int * sin
            pos_y_wf = y_int * cos - x_int * sin
            vel_x_int = vel_x_wf
            vel_y_int = vel_y_wf
            vel_x_wf = vel_x_int * cos + vel_y_int * sin
            vel_y_wf = vel_y_int * cos - vel_x_int * sin
            yaw = smallest_angle_diff_rad(yaw, self._flight_origin[2])

        cos = np.cos(yaw)
        sin = np.sin(yaw)
        vel_z_df = vel_z_wf
        vel_x_df = vel_x_wf * cos + vel_y_wf * sin
        vel_y_df = vel_y_wf * cos - vel_x_wf * sin

        if self.trace_logs:
            now = time.time() - self.start_time
            self.x_tracer.write_line(f"{now};{pos_x_wf};{vel_x_wf};{vel_x_df}")
            self.y_tracer.write_line(f"{now};{pos_y_wf};{vel_y_wf};{vel_y_df}")

        self.telemetry = (self.voltage, pos_x_wf, pos_y_wf, pos_z_wf, yaw, vel_x_wf, vel_y_wf, vel_z_wf, yaw_rate, self.debug_flags)

        if self.current_flight_command is None:
            # ensure that rpy commands are at default:
            self.CMDS['pitch'] = DEFAULT_PITCH
            self.CMDS['roll'] = DEFAULT_ROLL
            self.CMDS['yaw'] = DEFAULT_YAW
        else:
            # handle the current command
            if self.current_flight_command[0] in ("VDF", "VWF"):  # velocity drone/world frame command
                # command is ["VXF", x, y, z, time_end]
                time_end = self.current_flight_command[5]
                if time.time() >= time_end:
                    # stop moving and remove flight command
                    self.pid_vel_x.set_auto_mode(False)
                    self.pid_vel_y.set_auto_mode(False)
                    self.pid_vel_z.set_auto_mode(False)
                    self.pid_w_z.set_auto_mode(False)
                    self.CMDS['pitch'] = DEFAULT_PITCH
                    self.CMDS['roll'] = DEFAULT_ROLL
                    self.CMDS['yaw'] = DEFAULT_YAW
                    self.current_flight_command = None
                else:
                    if self.current_flight_command[0] == "VDF":  # drone frame
                        v_x = self.current_flight_command[1]
                        v_y = self.current_flight_command[2]
                    else:  # world frame
                        v_x_wf = self.current_flight_command[1]
                        v_y_wf = self.current_flight_command[2]
                        v_x = v_x_wf * cos + v_y_wf * sin
                        v_y = v_y_wf * cos - v_x_wf * sin
                    v_z = self.current_flight_command[3]
                    w = self.current_flight_command[4]
                    # adapt pitch, roll and throttle
                    self.pid_vel_x.setpoint = v_x
                    self.pid_vel_y.setpoint = v_y
                    self.pid_vel_z.setpoint = v_z
                    self.pid_w_z.setpoint = w
                    self.pid_vel_x.set_auto_mode(True)
                    self.pid_vel_y.set_auto_mode(True)
                    if v_z != 0:
                        self.pid_vel_z.set_auto_mode(True)
                    else:
                        self.pid_vel_z.set_auto_mode(False)
                    if w != 0:
                        self.pid_w_z.set_auto_mode(True)
                    else:
                        self.pid_w_z.set_auto_mode(False)
                    x_target = self.pid_vel_x(vel_x_df)
                    y_target = self.pid_vel_y(vel_y_df)
                    z_target = self.pid_vel_z(vel_z_df) if v_z != 0 else 0
                    w_target = self.pid_w_z(yaw_rate) if w != 0 else 0
                    self.CMDS['pitch'] = DEFAULT_PITCH + x_target
                    self.CMDS['roll'] = DEFAULT_ROLL + y_target
                    self.CMDS['throttle'] = self.CMDS['throttle'] + z_target
                    self.CMDS['yaw'] = DEFAULT_YAW + w_target
            elif self.current_flight_command[0] == "PWF":  # position command
                # command is ["PWF", x, y, z, yaw, vel_norm_goal, w_norm_goal, duration]
                time_end = self.current_flight_command[7]
                if time.time() >= time_end:
                    # stop moving and remove flight command
                    self.pid_vel_x.set_auto_mode(False)
                    self.pid_vel_y.set_auto_mode(False)
                    self.pid_vel_z.set_auto_mode(False)
                    self.pid_w_z.set_auto_mode(False)
                    self.CMDS['pitch'] = DEFAULT_PITCH
                    self.CMDS['roll'] = DEFAULT_ROLL
                    self.CMDS['yaw'] = DEFAULT_YAW
                    self.current_flight_command = None
                else:
                    # compute a 4D vector toward the goal:
                    x_goal = self.current_flight_command[1]
                    y_goal = self.current_flight_command[2]
                    z_goal = self.current_flight_command[3]
                    yaw_goal = self.current_flight_command[4]
                    vel_norm_goal = self.current_flight_command[5]
                    w_norm_goal = self.current_flight_command[6]
                    x_vector = x_goal - pos_x_wf
                    y_vector = y_goal - pos_y_wf
                    z_vector = z_goal - pos_z_wf
                    yaw_vector = smallest_angle_diff_rad(yaw_goal, yaw) if yaw_goal is not None else None
                    if self.target_flag:  # check whether position target is reached
                        dist_condition = np.linalg.norm([x_vector, y_vector, z_vector]) <= EPSILON_DIST_TO_TARGET
                        angle_condition = True if yaw_vector is None else abs(yaw_vector) <= EPSILON_ANGLE_TO_TARGET
                        if dist_condition and angle_condition:
                            self.target_flag = False  # disable the flag
                            if self.sender_initialized:
                                self.udp_int.send(pkl.dumps(("DON", self.target_id)))  # notify the remote control
                    # convert x and y to the drone frame:
                    x_vector_df = x_vector * cos + y_vector * sin
                    y_vector_df = y_vector * cos - x_vector * sin
                    # compute the corresponding desired velocity (non-clipped):
                    v_x = self.x_vel_gain * x_vector_df
                    v_y = self.y_vel_gain * y_vector_df
                    v_z = self.z_vel_gain * z_vector
                    w = self.w_gain * yaw_vector if yaw_goal is not None else 0.0
                    # clip to desired velocty:
                    vel_norm = np.linalg.norm([v_x, v_y, v_z])
                    w_norm = abs(w)
                    if vel_norm > vel_norm_goal:
                        v_x = v_x * vel_norm_goal / vel_norm
                        v_y = v_y * vel_norm_goal / vel_norm
                        v_z = v_z * vel_norm_goal / vel_norm
                    if w_norm > w_norm_goal:
                        w = w * w_norm_goal / w_norm
                    # apply the desired velocity:
                    self.pid_vel_x.setpoint = v_x
                    self.pid_vel_y.setpoint = v_y
                    self.pid_vel_z.setpoint = v_z
                    self.pid_w_z.setpoint = w
                    self.pid_vel_x.set_auto_mode(True)
                    self.pid_vel_y.set_auto_mode(True)
                    if v_z != 0:
                        self.pid_vel_z.set_auto_mode(True)
                    else:
                        self.pid_vel_z.set_auto_mode(False)
                    if w != 0:
                        self.pid_w_z.set_auto_mode(True)
                    else:
                        self.pid_w_z.set_auto_mode(False)
                    x_target = self.pid_vel_x(vel_x_df)
                    y_target = self.pid_vel_y(vel_y_df)
                    z_target = self.pid_vel_z(vel_z_df) if v_z != 0 else 0
                    w_target = self.pid_w_z(yaw_rate) if w != 0 else 0
                    self.CMDS['pitch'] = DEFAULT_PITCH + x_target
                    self.CMDS['roll'] = DEFAULT_ROLL + y_target
                    self.CMDS['throttle'] = self.CMDS['throttle'] + z_target
                    self.CMDS['yaw'] = DEFAULT_YAW + w_target
            elif self.current_flight_command[0] == "PDF":  # position command drone frame
                # command is ["PDF", x, y, z, yaw, vel_norm_goal, w_norm_goal, duration]
                # we convert this into a command in world frame (it will be applied in the next iteration)
                current_flight_command = list(self.current_flight_command)
                current_flight_command[0] = "PWF"
                x_goal_df = current_flight_command[1]
                y_goal_df = current_flight_command[2]
                x_goal_wf = x_goal_df * cos - y_goal_df * sin + pos_x_wf
                y_goal_wf = y_goal_df * cos + x_goal_df * sin + pos_y_wf
                yaw_goal_df = current_flight_command[4]
                yaw_goal_wf = (yaw + yaw_goal_df) % (2 * np.pi) if yaw_goal_df is not None else None
                current_flight_command[1] = x_goal_wf
                current_flight_command[2] = y_goal_wf
                current_flight_command[4] = yaw_goal_wf
                self.current_flight_command = tuple(current_flight_command)

    def _controller(self):
        try:
            average_cycle = deque([0.0] * NO_OF_CYCLES_AVERAGE_GUI_TIME)

            last_loop_time = last_slow_msg_time = last_cycle_time = time.time()
            while True:
                start_time = time.time()

                #
                # UDP recv non-blocking  (NO DELAYS) -----------------------
                # For safety, UDP commands are overridden by key presses
                #
                if self.udp_int:
                    udp_cmds = self.udp_int.recv_nonblocking()
                    if len(udp_cmds) > 0:
                        for cmd in udp_cmds:
                            self._udp_commands_handler(pkl.loads(cmd))
                    self._flight()
                    if self.obs_loop_time is not None:
                        tick = time.time()
                        if tick - self.last_obs_tick >= self.obs_loop_time and self.sender_initialized:
                            self.last_obs_tick = tick
                            self._i_obs += 1
                            self.udp_int.send(pkl.dumps(("OBS", self._i_obs, self.telemetry)))
                #
                # end of UDP recv non-blocking -----------------------------
                #

                #
                # CLIP RC VALUES -------------------------------------------
                #
                self.CMDS['roll'] = clip(self.CMDS['roll'], MIN_CMD_ROLL, MAX_CMD_ROLL)
                self.CMDS['pitch'] = clip(self.CMDS['pitch'], MIN_CMD_PITCH, MAX_CMD_PITCH)
                self.CMDS['yaw'] = clip(self.CMDS['yaw'], MIN_CMD_YAW, MAX_CMD_YAW)
                self.CMDS['throttle'] = clip(self.CMDS['throttle'], MIN_CMD_THROTTLE, MAX_CMD_THROTTLE)
                #
                # END CLIP RC VALUES ---------------------------------------
                #

                # This slows the loop down to CTRL_LOOP_TIME:
                end_time = time.time()
                last_cycle_time = end_time - start_time
                if last_cycle_time < CTRL_LOOP_TIME:
                    time.sleep(CTRL_LOOP_TIME - last_cycle_time)

                average_cycle.append(last_cycle_time)
                average_cycle.popleft()

        finally:
            print("Bye!")


def run_controller():
    cc = CogniflyControllerDummy()
    cc.run_curses()


if __name__ == "__main__":
    run_controller()



"""
Usage:
  python3 -m cognifly controller
  or
  python3 -m cognifly controller-headless
"""

import sys
from cognifly.cognifly_controller.cognifly_controller import run_controller

_, cmd, *args = sys.argv

if cmd == "controller":
    run_controller()
elif cmd == "controller-headless":
    run_controller(print_screen=False)
else:
    raise AttributeError("Undefined command: " + cmd)



import socket
import logging
import platform


WINDOWS = platform.system() == 'Windows'


class UDPInterface(object):
    def __init__(self):
        self.ip_send = None
        self.port_send = None
        self.ip_recv = None
        self.port_recv = None
        self.__buffersize = None
        self.__sockO = None
        self.__sockI = None

    def init_sender(self, ip, port):
        if self.__sockO is not None:
            self.__sockO.close()
            self.__sockO = None
        self.__sockO = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.ip_send = ip
        self.port_send = port

    def init_receiver(self, ip, port, clean_buffer=True):
        self.ip_recv = ip
        self.port_recv = port
        if self.__sockI is None:
            self.__sockI = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        else:
            raise Exception("Cannot initialize the receiver twice.")
        self.__sockI.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.__sockI.bind((ip, port))
        self.__buffersize = self.__sockI.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)
        logging.debug(f"buffer size: {self.__buffersize}")
        while clean_buffer:
            logging.debug(f"Cleaning receiving buffer...")
            try:
                if not WINDOWS:
                    self.__sockI.recv(1, socket.MSG_DONTWAIT)
                else:
                    self.__sockI.setblocking(False)
                    self.__sockI.recv(1)
            except IOError:  # recv raises a error when no data is received
                clean_buffer = False
            finally:
                if WINDOWS:
                    self.__sockI.setblocking(True)
        logging.debug(f"Cleaning receiving buffer...Done!")

    def __del__(self):
        if self.__sockI:
            self.__sockI.close()
        if self.__sockO:
            self.__sockO.close()

    def send(self, data):
        """
        data: binary string
        """
        assert self.__sockO is not None, "sender not initialized"

        self.__sockO.sendto(data, (self.ip_send, self.port_send))

    def recv(self, timeout=None):
        """
        This returns a list of bytestrings (all messages in the buffer)
        Returns an empty list if no message is present in the buffer
        Reads self.__buffersize bytes in the buffer
        """
        assert self.__sockI is not None, "receiver not initialized"

        if timeout:
            self.__sockI.settimeout(timeout)
        try:
            data, _ = self.__sockI.recvfrom(self.__buffersize)
            self.__sockI.settimeout(None)
        except socket.timeout:
            return []
        res = [data]
        nb = self.recv_nonblocking()
        res = res + nb
        return res

    def recv_nonblocking(self):
        """
        This returns a list of bytestrings (all messages in the buffer)
        Returns an empty list if no message is present in the buffer
        Reads everything in the buffer
        """
        assert self.__sockI is not None, "receiver not initialized"

        res = []
        while True:
            try:
                if not WINDOWS:
                    data, _ = self.__sockI.recvfrom(self.__buffersize, socket.MSG_DONTWAIT)
                else:
                    self.__sockI.setblocking(False)
                    data, _ = self.__sockI.recvfrom(self.__buffersize)
            except IOError:
                return res
            finally:
                if WINDOWS:
                    self.__sockI.setblocking(True)
            res += [data]


def test_send(interface, msg):
    print(f"sending: {msg}")
    interface.send(msg)


def test_recv(interface, timeout=None):
    print(f"receiving (blocking)...")
    res = interface.recv(timeout)
    print(f"received: {res}")


def test_recv_nonblocking(interface):
    print(f"receiving (non blocking)...")
    res = interface.recv_nonblocking()
    print(f"received: {res}")


def main(args):
    # standard library imports
    import time

    ip_send = args.ipsend
    ip_recv = args.iprecv
    port_send = args.portsend
    port_recv = args.portrecv

    conn = UDPInterface()
    conn.init_sender(ip_send, port_send)
    conn.init_receiver(ip_recv, port_recv)

    test_send(conn, b"hello 0")
    test_send(conn, b"hello 1")
    test_recv(conn)
    test_send(conn, b"hello 2")
    test_send(conn, b"hello 3")
    test_recv_nonblocking(conn)
    test_recv_nonblocking(conn)
    test_send(conn, b"hello 4")
    test_recv(conn, timeout=1.0)
    test_recv(conn, timeout=1.0)


if __name__ == "__main__":
    # standard library imports
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--ipsend', type=str, default="127.0.0.1", help='IP address of the drone if any.')
    parser.add_argument('--iprecv', type=str, default="127.0.0.1", help='local IP address if any.')
    parser.add_argument('--portsend', type=int, default=8989, help='Port to send udp messages to.')
    parser.add_argument('--portrecv', type=int, default=8989, help='Port to reveive udp messages from.')
    args = parser.parse_args()
    main(args)



# code inspired from https://picamera.readthedocs.io/en/release-1.13/recipes2.html

import io
import socket
import struct
import time
import logging
from PIL import Image
from threading import Thread, Lock, Condition
from copy import deepcopy
import numpy as np
import traceback
from cognifly.utils.ip_tools import get_free_port


class SplitFrames(object):
    def __init__(self):  # , connection
        # self.connection = connection
        self.stream = io.BytesIO()
        self.count = 0
        self._condition_out = Condition()
        self.__frame = None
        self.__framelen = None

    def write(self, buf):
        if buf.startswith(b'\xff\xd8'):
            # Start of new frame; send the old one's length
            # then the data
            size = self.stream.tell()
            if size > 0:
                with self._condition_out:
                    self.__framelen = size
                    # self.connection.write(struct.pack('<L', size))
                    # self.connection.flush()
                    self.stream.seek(0)
                    # self.connection.write(self.stream.read(size))
                    self.__frame = self.stream.read(size)
                    self._condition_out.notifyAll()
                self.count += 1
                self.stream.seek(0)
        self.stream.write(buf)

    def get_frame(self):
        return self.__framelen, self.__frame


class TCPVideoInterface(object):
    def __init__(self, camera=None):
        self.ip_send = None
        self.port_send = None
        self.__buffersize = None
        self.__sockO = None
        self.__lock = Lock()
        self.__record = False
        self.__stream_running = False
        self.__receiver_running = False
        self.__camera_error = False
        self.__camera_exception = None
        self.__camera_traceback = None
        self.__image = None
        self.__image_i = -1
        self.__receive = False
        self.condition_in = Condition()
        self.camera = camera

    def get_camera_error(self):
        with self.__lock:
            err = self.__camera_error
            exc = self.__camera_exception
            trace = self.__camera_traceback
        return err, exc, trace

    def __streaming_thread(self, ip, port, wait_duration, resolution, fps):
        camera_error = False
        connection = None
        client_socket = None
        camera = None
        try:
            import picamera  # can be imported only on the raspberry pi
            client_socket = socket.socket()
            client_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            client_socket.connect((ip, port))
            connection = client_socket.makefile('wb')
            output = SplitFrames()  # connection
            with picamera.PiCamera(resolution=resolution, framerate=fps) as camera:
                time.sleep(2)
                with self.__lock:
                    self.__record = True
                    record = True
                camera.start_recording(output, format='mjpeg')
                while record:
                    # retrieve freshest frame
                    with output._condition_out:
                        output._condition_out.wait()
                        framelen, frame = output.get_frame()
                    # send length and frame:
                    connection.write(struct.pack('<L', framelen))
                    connection.flush()
                    connection.write(frame)
                    # camera.wait_recording(wait_duration)
                    # check whether we must keep recording:
                    with self.__lock:
                        record = self.__record
                camera.stop_recording()
                # Write the terminating 0-length to the connection to let the
                # server know we're done
                connection.write(struct.pack('<L', 0))
            if connection is not None:
                connection.close()
            if client_socket is not None:
                client_socket.close()
        except Exception as e:
            logging.info("error")
            t_e = type(e)
            if t_e != OSError and not issubclass(t_e, ConnectionError):
                with self.__lock:
                    self.__camera_error = True
                    self.__camera_exception = str(e)
                    self.__camera_traceback = ''.join(traceback.format_tb(e.__traceback__))
                # raise e
        finally:
            if camera is not None:
                camera.close()
            with self.__lock:
                self.__stream_running = False
                self.__record = False

    def start_streaming(self, ip_dest, port_dest, wait_duration=1.0, resolution='VGA', fps=10):
        """
        Starts the streaming thread on the drone.
        Caution: stop_streaming() must be called between calls to start_streaming().
        """
        with self.__lock:
            streaming_started = self.__stream_running
            if not streaming_started:
                self.__stream_running = True
        if not streaming_started:
            start_streaming_thread = Thread(target=self.__streaming_thread, args=(ip_dest, port_dest, wait_duration, resolution, fps))
            start_streaming_thread.setDaemon(True)  # thread will be terminated at exit
            start_streaming_thread.start()
        else:
            logging.info("start_streaming() called although streaming is already started.")

    def stop_streaming(self):
        """
        Ends the streaming thread on the drone
        """
        with self.__lock:
            self.__record = False

    def __stream_receiver_thread(self, port):  # display=False
        """
        TCP server, should be running prior to the client
        """
        # if display:
        #     import cv2

        server_socket = socket.socket()
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind(('0.0.0.0', port))
        server_socket.listen(0)

        # Accept a single connection and make a file-like object out of it
        connection = server_socket.accept()[0].makefile('rb')
        try:
            while True:
                # Read the length of the image as a 32-bit unsigned int. If the
                # length is zero, quit the loop
                image_len = struct.unpack('<L', connection.read(struct.calcsize('<L')))[0]
                if not image_len:
                    break
                # Construct a stream to hold the image data and read the image
                # data from the connection
                image_stream = io.BytesIO()
                image_stream.write(connection.read(image_len))
                # Rewind the stream, open it as an image with PIL and do some
                # processing on it
                image_stream.seek(0)
                image = Image.open(image_stream)

                # if display:
                #     open_cv_image = np.array(image)
                #     im = open_cv_image[:, :, ::-1].copy()
                #     cv2.imshow("Stream", im)
                #     cv2.waitKey(1)

                with self.condition_in:
                    with self.__lock:
                        self.__image_i += 1
                        self.__image = image
                        receive = self.__receive
                    self.condition_in.notifyAll()
                if not receive:
                    break
        finally:
            # if display:
            #     cv2.destroyAllWindows()
            connection.close()
            server_socket.close()
            with self.__lock:
                self.__receiver_running = False

    def is_receiver_running(self):
        with self.__lock:
            res = self.__receiver_running
        return res

    def start_receiver(self, port):
        with self.__lock:
            self.__receive = True
            receiver_running = self.__receiver_running
            self.__receiver_running = True
        if not receiver_running:
            receiver_thread = Thread(target=self.__stream_receiver_thread, args=(port, ))
            receiver_thread.setDaemon(True)  # thread will be terminated at exit
            receiver_thread.start()

    def stop_receiver(self):
        with self.__lock:
            self.__receive = False
            receiver_running = self.__receiver_running
        while receiver_running:
            with self.__lock:
                receiver_running = self.__receiver_running
            if receiver_running:
                time.sleep(0.01)

    def get_last_image(self, min_image_i=-1):
        """
        Outputs the last image and its identifier when it is more recent than the image with the inf_image_i identifier
        Args:
            min_image_i: int: identifier of the previously received image (starts at 0 and increases)
        Returns:
            im: PIL.Image or None: None when the current identifier is not > inf_image_i
            im_i: int: identifier
        """
        with self.__lock:
            im_i = self.__image_i
            data = deepcopy(self.__image) if im_i > min_image_i else None
        if data is not None:
            im = data
            pil_image = data.convert('RGB')
            open_cv_image = np.array(pil_image)
            im = open_cv_image[:, :, ::-1].copy()
        else:
            im = None
        return im, im_i

    def __del__(self):
        if self.__sockO:
            try:
                self.__sockO.close()
            except BrokenPipeError:
                pass
        self.stop_receiver()



import time
import numpy as np


class Simple1DKalman:
    def __init__(self, err_measure, err_estimate=None, initial_estimate=0, q=0.01):
        """
        Args:
            err_measure: float: error that can be expected in the measurements (important)
            err_estimate: float (optional): error in the initial estimate (if None, same as err_measure)
            initial_estimate: float (optional): initial estimate
            q: float (optional): process variance (between 0 and 1, depending on how fast the process moves)
        """
        self._err_measure = err_measure
        self._err_estimate = err_estimate if err_estimate is not None else err_measure
        self._last_estimate = initial_estimate
        self._q = q

    def update_estimate(self, mea):
        kalman_gain = self._err_estimate / (self._err_estimate + self._err_measure)
        current_estimate = self._last_estimate + kalman_gain * (mea - self._last_estimate)
        self._err_estimate = (1.0 - kalman_gain) * self._err_estimate + self._q * abs(self._last_estimate - current_estimate)
        self._last_estimate = current_estimate
        return current_estimate


class Simple1DExponentialAverage:
    def __init__(self, tau, initial_estimate=0):
        """
        Args:
            tau: float: > 0, time constant of the filter
            initial_estimate: float (optional): initial estimate
        """
        self.tau = tau
        self._last_estimate = initial_estimate
        self._last_ts = time.monotonic()

    def _get_alpha(self):
        now = time.monotonic()
        dt = now - self._last_ts
        alpha = 1 - np.exp(- dt / self.tau)
        self._last_ts = now
        return alpha

    def update_estimate(self, mea):
        alpha = self._get_alpha()
        current_estimate = alpha * mea + (1 - alpha) * self._last_estimate
        self._last_estimate = current_estimate
        return current_estimate



import numpy as np

def clip(x, min_x, max_x):
    """
    clips x between min_x and max_x
    """
    return max(min(x, max_x), min_x)


def get_angle_sequence_rad(angle):
    """
    cuts an angle bigger than pi into a sequence of angles <= pi/2
    """
    aa = abs(angle)
    if aa >= np.pi:  # equality is important because the direction is undefined
        chunks = int(aa // (np.pi / 2.0))
        last = aa % (np.pi / 2.0)
        res = [np.pi / 2.0, ] * chunks
        if last != 0:
            res.append(last)
    else:
        res = [aa, ]
    res = np.array(res) * np.sign(angle)
    return res.tolist()


def smallest_angle_diff_rad(theta_new, theta_old):
    """
    computes theta_new - theta_old (radian)
    """
    diff = theta_new - theta_old
    mod_p = diff % (2 * np.pi)
    mod_n = diff % (-2 * np.pi)
    return mod_p if abs(mod_p) < abs(mod_n) else mod_n



import sys
import os
import io
from setuptools import setup, find_packages

if sys.version_info < (3, 7):
    sys.exit('Sorry, Python < 3.7 is not supported.')


def is_raspberrypi():
    try:
        with io.open('/sys/firmware/devicetree/base/model', 'r') as m:
            if 'raspberry pi' in m.read().lower():
                return True
    except Exception:
        pass
    return False


with open("README.md", "r") as fh:
    long_description = fh.read()

deps = ['numpy', 'Pillow', 'psutil']
if is_raspberrypi():  # drone
    deps.append(['picamera', 'yamspy'])
else:  # remote
    deps.append(['opencv-contrib-python', 'pysimplegui'])

setup(name='cognifly',
      packages=[package for package in find_packages()],
      version='0.1.1',
      license='MIT',
      description='Control the CogniFly open-source drone from python',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Yann Bouteiller',
      url='https://github.com/thecognifly/cognifly-python',
      download_url='https://github.com/thecognifly/cognifly-python/archive/refs/tags/v0.1.1.tar.gz',
      keywords=['cognifly', 'drone', 'remote', 'control'],
      install_requires=deps,
      scripts=["scripts/cognifly-controller", ],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Information Technology',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python',
          'Framework :: Robot Framework :: Library',
          'Topic :: Education',
          'Topic :: Scientific/Engineering :: Artificial Intelligence',
      ],
      )



from cognifly.cognifly_remote.cognifly_remote import Cognifly

import logging

logfilename = "cognifly.log"
logfilemode = 'a'
loglevel = 'INFO'

logging.basicConfig(format="[%(levelname)s] [%(asctime)s]: %(message)s",
                    filename=logfilename,
                    filemode=logfilemode,
                    level=getattr(logging, loglevel.upper()))

